"""
Collection of plotting functions
"""
import matplotlib.pyplot as plt
import numpy as np
def plot_spectrum(axis, spectrum, units='cm-1', output_name = None, fig_size=(10,8), **kwargs):
    """
    Plot Spectrum with Luci format. If output name is supplied, the plot will be saved
    Args:
        axis: X axis of spectrum (1d numpy array)
        spectrum: Y axis of spectrum (1d numpy array)
        units: Wavelength units (e.x. 'cm')
        output_name: Path to output file (default None)
    """
    f, ax = plt.subplots(figsize=fig_size)
    check_units(units)  # Check that user supplied appropriate wavelength option
    if units == 'nm':
        axis = [1e7/axis_val for axis_val in axis]
        ax.set_xlim(635, 675)
    else:
        ax.set_xlim(14750, 15750)
    ax.plot(axis, spectrum, **kwargs)
    ax.set_xlabel(r"Wavelength [%s]"%units, fontsize=20, fontweight='bold')
    ax.set_ylabel(r'Flux [ergs s$^{-1}$ cm$^{-2}$ $\AA^{-1}$]', fontsize=20, fontweight='bold')
    ax.tick_params(labelsize=14)
    if output_name is not None:
        plt.savefig(output_name)
    return ax


def plot_map(quantity_map, quantity_name, output_dir, clims=None):
    """
    Function to plot fit map
    Args:
        quantity_map: 2d numpy array from fit
        quantity_name: Name of quantity (e.x. 'flux')
        output_dit: Path (absolute or partial) to output directory
        clims: List containing lower and upper limits of colorbar (e.x. [-500, 500])
    """
    if quantity_name == 'broadening' or quantity_name == 'velocity':
        quantity_name = 'velocity'  # The quantities are the same
    elif quantity_name != 'flux':
        quantity_map = np.log10(quantity_map)
        print('Please enter either flux, velocity, or broadening')
    units = {'flux':'ergs/s/cm^2/A','velocity':'km/s'}
    if clims is None:
        c_min = np.nanpercentile(quantity_map, 5)
        c_max = np.nanpercentile(quantity_map, 95)
    else:
        c_min = clims[0]
        c_max = clims[1]
    #Plot
    #hdu = fits.open(Name+'_SN3.1.0.ORCS/MAPS/'+Name+'_SN3.1.0.LineMaps.map.all.'+Bin+'.rchi2.fits')[0]
    #wcs = WCS(hdu.header)
    fig = plt.figure(figsize=(8,10))
    ax = plt.subplot()#projection=wcs)
    #ax.coords[0].set_major_formatter('hh:mm:ss')
    #ax.coords[1].set_major_formatter('dd:mm:ss')
    plt.imshow(quantity_map, cmap='jet')
    plt.title((quantity_name +' map').upper(), fontsize=26, fontweight='bold')
    plt.xlabel("RA", fontsize=20, fontweight='bold')
    plt.ylabel("DEC", fontsize=20, fontweight='bold')
    plt.xlim(0,quantity_map.shape[0])
    plt.ylim(0,quantity_map.shape[1])
    cbar = plt.colorbar(fraction=0.046, pad=0.04)
    plt.clim(c_min, c_max)
    cbar.ax.set_ylabel(units[quantity_name], rotation=270, labelpad=25, fontsize=20, fontweight='bold')
    plt.savefig(output_dir+'/'+quantity_name+'_map.png')


def check_units(unit):
    """
    This function checks to see that the unit provided is in the available options
    Args:
        unit: User supplied unit
    Return:
        Nothing if the user provides an appropriate unit
        Else it will throw an error
    """
    unit_options = ['cm-1', 'nm']
    if unit in unit_options:
        pass
    else:
        raise Exception('Please submit a unit name in the available list: \n {}'.format(unit_options))



"""
Suite of tests for Luci
"""
import numpy as np
from astropy.io import fits
import keras

from LuciFit import Fit
from LuciFit import Gaussian


class Test:
    def __init__(self):
        self.amp = 1
        self.pos = 15234
        self.sigma = 2.5
        self.model_type = 'gaussian'
        self.lines = ['Halpha']
        self.ML_model = None  # keras.models.load_model('ML/R5000-PREDICTOR-I')
        self.axis = None
        self.spectrum = None
        self.wavenumbers_syn = None
        self.make_spectrum()
        self.read_ref_spec()
        self.LuciFit_ = Fit(self.spectrum, self.axis, self.wavenumbers_syn, self.model_type, self.lines, self.ML_model)

    def make_spectrum(self):
        self.axis = np.linspace(14400, 15800, 10000)
        self.spectrum = Gaussian(self.axis, [self.amp, self.pos, self.sigma]).func
        return None

    def read_ref_spec(self):
        ref_spec = fits.open('ML/Reference-Spectrum-R5000.fits')[1].data
        channel = []
        counts = []
        for chan in ref_spec:  # Only want SN3 region
            channel.append(chan[0])
            counts.append(np.real(chan[1]))
        min_ = np.argmin(np.abs(np.array(channel) - 14700))
        max_ = np.argmin(np.abs(np.array(channel) - 15600))
        self.wavenumbers_syn = np.array(channel[min_:max_], dtype=np.float32)


    def test_spec_length(self):
        assert 'Halpha' in self.lines

    def test_fit(self):
        self.LuciFit_.fit()
        # Check that fit is within 10%
        assert self.LuciFit_.fit_sol[0]-1 < 0.1


def test():
    Test_ = Test()
    Test_.test_spec_length()
    Test_.test_fit()


import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize
from scipy import interpolate
import keras
from scipy.optimize import Bounds
from numdifftools import Jacobian, Hessian
import emcee
from scipy.stats import chisquare

import warnings

warnings.filterwarnings("ignore")


class Gaussian:
    def __init__(self, channel, params):
        A = params[0];
        x = params[1];
        sigma = params[2]
        self.func = A * np.exp((-(channel - x) ** 2) / (2 * sigma ** 2))


class Sinc:
    def __init__(self, channel, params):
        p0 = params[0];
        p1 = params[1];
        p2 = params[2]
        u = (channel - p1) / p2
        self.func = p0 * (np.sin(u) / u)


class Fit:
    """
    Class that defines the functions necessary for the modelling aspect. This includes
    the gaussian fit functions, the prior definitions, the log likelihood, and the
    definition of the posterior (log likelihood times prior).

    The initial arguments are as follows:
    Args:

        spectrum: Spectrum of interest. This should not be the interpolated spectrum nor normalized(numpy array)

        axis: Wavelength Axis of Spectrum (numpy array)

        wavenumbers_syn: Wavelength Axis of Reference Spectrum (numpy array)

        model_type: Type of model ('gaussian')

        lines: Lines to fit (must be in line_dict)

        ML_model: Tensorflow/keras machine learning model

        Plot_bool: Boolean to determine whether or not to plot the spectrum (default = False)

    """

    def __init__(self, spectrum, axis, wavenumbers_syn, model_type, lines,
                 ML_model, bayes_bool=False, Plot_bool=False):
        """
        Args:
            spectrum: Spectrum of interest. This should not be the interpolated spectrum nor normalized(numpy array)
            axis: Wavelength Axis of Spectrum (numpy array)
            wavenumbers_syn: Wavelength Axis of Reference Spectrum (numpy array)
            model_type: Type of model ('gaussian')
            lines: Lines to fit (must be in line_dict)
            ML_model: Tensorflow/keras machine learning model
            bayes_bool:
            Plot_bool: Boolean to determine whether or not to plot the spectrum (default = False)

        """
        self.line_dict = {'Halpha': 656.280, 'NII6583': 658.341, 'NII6548': 654.803,
                          'SII6716': 671.647, 'SII6731': 673.085, 'OII3726': 372.603,
                          'OII3729': 372.882, 'OIII4959': 495.891, 'OIII5007': 500.684,
                          'Hbeta': 486.133}
        self.available_functions = ['gaussian']
        self.spectrum = spectrum
        self.axis = axis
        self.wavenumbers_syn = wavenumbers_syn
        self.model_type = model_type
        self.lines = lines
        self.line_num = len(lines)  # Number of  lines to fit
        self.spectrum_interpolated = np.zeros_like(self.spectrum)
        self.spectrum_normalized = self.spectrum / np.max(self.spectrum)  # Normalized spectrum
        self.spectrum_interp_norm = np.zeros_like(self.spectrum)
        # ADD ML_MODEL AND PLOT_BOOL
        self.ML_model = ML_model
        self.bayes_bool = bayes_bool
        self.Plot_bool = Plot_bool
        self.spectrum_scale = 0.0  # Sacling factor used to normalize spectrum
        self.vel_ml = 0.0  # ML Estimate of the velocity [km/s]
        self.broad_ml = 0.0  # ML Estimate of the velocity dispersion [km/s]
        self.fit_sol = np.zeros(3 * self.line_num)  # Solution to the fit
        # Set bounds
        self.A_min = 0;
        self.A_max = 1.1;
        self.x_min = 14700;
        self.x_max = 15600
        self.sigma_min = 0;
        self.sigma_max = 10

        # Check that lines inputted by user are in line_dict
        self.check_lines()
        self.check_fitting_model()

    def estimate_priors_ML(self):
        """
        TODO: Test
        Apply machine learning algorithm on spectrum in order to estimate the velocity.
        The spectrum fed into this method must be interpolated already onto the
        reference spectrum axis AND normalized as described in Rhea et al. 2020a.
        Args:
            ml_dir: Relative path to the trained ML Predictor (e.g. R5000-PREDICITOR-I)
        Return:
            Updates self.vel_ml
        """
        Spectrum = self.spectrum_interp_norm.reshape(1, self.spectrum_interp_norm.shape[0], 1)
        predictions = self.ML_model(Spectrum, training=False)
        self.vel_ml = float(predictions[0][0])
        self.broad_ml = float(predictions[0][1])  # Multiply value by FWHM of a gaussian
        return None

    def interpolate_spectrum(self):
        """
        Interpolate Spectrum given the wavelength axis of reference spectrum.
        Then normalize the spectrum so that the max value equals 1

        Return:
            Populates self.spectrum_interpolated, self.spectrum_scale, and self.spectrum_interp_norm.

        """
        f = interpolate.interp1d(self.axis, self.spectrum, kind='slinear')
        self.spectrum_interpolated = f(self.wavenumbers_syn)
        self.spectrum_scale = np.max(self.spectrum_interpolated)
        self.spectrum_interp_norm = self.spectrum_interpolated / self.spectrum_scale
        # self.spectrum_interpolated = np.real(sky_corr)
        return None

    def line_vals_estimate(self, line_name):
        """
        TODO: Test

        Function to estimate the position and amplitude of a given line.

        Args:
            spec: Spectrum flux values
            line_name: Name of model. Available options are 'Halpha', 'NII6548', 'NII6543', 'SII6716', 'SII6731'

        Return:
            Estimated line amplitude in units of cm-1 (line_amp_est) and estimate line position in units of cm-1 (line_pos_est)

        """
        line_theo = self.line_dict[line_name]
        if self.ML_model is None or self.model_type == '':
            max_flux = np.argmax(self.spectrum)
            self.vel_ml = np.abs(3e5 * ((1e7/self.axis[max_flux] - line_theo) / line_theo))
            self.broad_ml = 10.0  # Best for now
        else:
            pass  # vel_ml and broad_ml already set using ML algorithm
        line_pos_est = 1e7 / ((self.vel_ml / 3e5) * line_theo + line_theo)  # Estimate of position of line in cm-1
        line_ind = np.argmin(np.abs(np.array(self.axis) - line_pos_est))
        line_amp_est = np.max([self.spectrum_normalized[line_ind - 2], self.spectrum_normalized[line_ind - 1],
                               self.spectrum_normalized[line_ind], self.spectrum_normalized[line_ind + 1],
                               self.spectrum_normalized[line_ind + 2]])
        line_broad_est = (line_pos_est * self.broad_ml) / 3e5
        return line_amp_est, line_pos_est, line_broad_est

    def gaussian_model(self, channel, theta):
        """
        Function to initiate the correct number of models to fit

        Args:
            channel: Wavelength Axis in cm-1
            theta: List of parameters for all the models in the following order
                            [amplitude, line location, sigma]

        Return:
            Value of function given input parameters (theta)

        """
        f1 = 0.0
        for model_num in range(self.line_num):
            params = theta[model_num * 3:(model_num + 1) * 3]
            f1 += Gaussian(channel, params).func
        return f1

    def log_likelihood(self, theta, yerr):
        """
        Calculate log likelihood function evaluated given parameters on spectral axis

        Args:
            theta - List of parameters for all the models in the following order
                            [amplitude, line location, sigma]
            yerr: Error on Spectrum's flux values (default 1e-2)
        Return:
            Value of log likelihood

        """
        model = self.gaussian_model(self.axis, theta)
        sigma2 = yerr ** 2
        return -0.5 * np.sum((self.spectrum_normalized - model) ** 2 / sigma2 + np.log(2 * np.pi * sigma2))

    def fun_der(self, theta, yerr):
        return Jacobian(lambda theta: self.log_likelihood(theta, yerr))(theta).ravel()

    def calculate_params(self):
        """
        Calculate the amplitude, position, and sigma of the line. These values are
        calculated using the scipy.optimize.minimize function. This is called
        on the log likelood previously described. The minimization algorithm uses
        the SLSQP optimization implementation. We have applied standard bounds in order
        to speed up the fitting. We also apply the fit on the normalized spectrum.
        We then correct the flux by un-normalizing the spectrum.

        """
        nll = lambda *args: -self.log_likelihood(*args)
        initial = np.ones((3 * self.line_num))
        bounds_ = []
        for mod in range(self.line_num):
            val = 3 * mod + 1
            amp_est, vel_est, sigma_est = self.line_vals_estimate(self.lines[mod])
            initial[3 * mod] = amp_est
            initial[3 * mod + 1] = vel_est
            initial[3 * mod + 2] = sigma_est
            bounds_.append((self.A_min, self.A_max))
            bounds_.append((self.x_min, self.x_max))
            bounds_.append((self.sigma_min, self.sigma_max))
        bounds_l = [val[0] for val in bounds_]
        bounds_u = [val[1] for val in bounds_]
        bounds = Bounds(bounds_l, bounds_u)
        self.inital_values = initial
        if self.line_num == 5:
            cons = ({'type': 'eq',
                     'fun': lambda x: 3e5 * ((1e7 / x[4] - self.line_dict['NII6583']) / (1e7 / x[4])) - 3e5 * (
                             (1e7 / x[1] - self.line_dict['Halpha']) / (1e7 / x[1]))},
                    {'type': 'eq', 'fun': lambda x: x[2] - x[5]},
                    {'type': 'eq', 'fun': lambda x: x[5] - x[8]},
                    {'type': 'eq', 'fun': lambda x: x[5] - x[11]},
                    {'type': 'eq', 'fun': lambda x: x[5] - x[14]},
                    {'type': 'eq',
                     'fun': lambda x: 3e5 * ((1e7 / x[4] - self.line_dict['NII6583']) / (1e7 / x[4])) - 3e5 * (
                             (1e7 / x[7] - self.line_dict['NII6548']) / (1e7 / x[7]))},
                    {'type': 'eq',
                     'fun': lambda x: 3e5 * ((1e7 / x[4] - self.line_dict['NII6583']) / (1e7 / x[4])) - 3e5 * (
                             (1e7 / x[10] - self.line_dict['SII6716']) / (1e7 / x[10]))},
                    {'type': 'eq',
                     'fun': lambda x: 3e5 * ((1e7 / x[4] - self.line_dict['NII6583']) / (1e7 / x[4])) - 3e5 * (
                             (1e7 / x[13] - self.line_dict['SII6731']) / (1e7 / x[13]))})
        else:
            cons = ()
        soln = minimize(nll, initial, method='SLSQP', # jac=self.fun_der(),
                        options={'disp': False, 'maxiter': 100}, bounds=bounds, tol=1e-8,
                        args=(1e-2), constraints=cons)
        parameters = soln.x
        # We now must unscale the amplitude
        for i in range(self.line_num):
            parameters[i * 3] *= self.spectrum_scale
        self.fit_sol = parameters
        self.fit_vector = self.gaussian_model(self.axis, self.fit_sol)
        return None

    def calculate_vel(self):
        """
        Calculate velocity given the fit of Halpha
        TODO: Test
        TODO: Add other lines

        Return:
            Velocity of the Halpha line in units of km/s
        """
        l_calc = 1e7 / self.fit_sol[1]  # Halpha
        l_shift = (l_calc - 656.28) / l_calc
        v = 3e5 * l_shift
        return v

    def calculate_broad(self):
        """
        Calculate velocity dispersion given the fit of Halpha
        TODO: Test
        TODO: Add other lines

        Return:
            Velocity Dispersion of the Halpha line in units of km/s
        """
        broad = (3e5 * self.fit_sol[2]) / self.fit_sol[1]
        return broad

    def calculate_flux(self, line_amp, line_sigma):
        """
        Calculate flux value given fit of line
        TODO: Test

        Args:
            line_amp: Amplitude of the line (un-normalized)
            line_sigma: Sigma of the line fit
        Return:
            Flux of the provided line in units of erg/s/cm-2
        """
        flux = 0.0  # Initialize
        if self.model_type == 'gaussian':
            flux = np.sqrt(2 * np.pi) * line_amp * line_sigma
        elif self.model_type == 'sinc':
            flux = np.sqrt(np.pi) * line_amp * line_sigma
        return flux

    def fit(self):
        """
        Primary function call for a spectrum. This will estimate the velocity using
        our machine learning algorithm described in Rhea et al. 2020a. Then we will
        fit our lines using scipy.optimize.minimize.

        Return:
            dictionary of parameters returned by the fit. The dictionary has the following form:
            {"fit_vector": Fitted spectrum, "velocity": Velocity of the line in km/s (float),
            "broadening": Velocity Dispersion of the line in km/s (float)}
        """
        # Interpolate Spectrum
        self.interpolate_spectrum()
        # Estimate the priors using machine learning algorithm
        if self.ML_model != None:
            self.estimate_priors_ML()
        # Apply Fit
        self.calculate_params()
        # Check if Bayesian approach is required
        if self.bayes_bool == True:
            self.fit_Bayes()
        # Calculate fit statistic
        chi_sqr, p_val = chisquare(self.fit_vector, self.spectrum)
        # Collect Amplitudes
        ampls = []
        fluxes = []
        for line_ct, line_ in enumerate(self.lines):  # Step through each line
            ampls.append(self.fit_sol[line_ct * 3])
            # Calculate flux
            fluxes.append(self.calculate_flux(self.fit_sol[line_ct * 3], self.fit_sol[line_ct * 3 + 2]))

        # Collect parameters to return in a dictionary
        fit_dict = {'fit_sol': self.fit_vector, 'velocity': self.calculate_vel(),
                    'broadening': self.calculate_broad(), 'amplitudes': ampls,
                    'fluxes': fluxes, 'chi2': chi_sqr}
        # Plot
        if self.Plot_bool == True:
            self.plot()
        return fit_dict

    def fit_Bayes(self):
        """
        Apply Bayesian MCMC run to constrain the parameters after solving
        """
        # Unscale the amplitude
        for i in range(self.line_num):
            self.fit_sol[i * 3] /= self.spectrum_scale
        n_dim = 3 * self.line_num
        n_walkers = n_dim * 2 + 4
        init_ = self.fit_sol + 1 * np.random.randn(n_walkers, n_dim)
        sampler = emcee.EnsembleSampler(n_walkers, n_dim, self.log_probability,
                                        args=(self.axis, self.spectrum_normalized, 1e-2, self.lines))
        sampler.run_mcmc(init_, 100, progress=False)
        flat_samples = sampler.get_chain(discard=20, flat=True)
        parameters = []
        for i in range(n_dim):
            mcmc = np.percentile(flat_samples[:, i], [16, 50, 84])
            parameters.append(mcmc[1])
        self.fit_sol = parameters
        # Now rescale the amplitude
        for i in range(self.line_num):
            self.fit_sol[i * 3] *= self.spectrum_scale

    def log_likelihood_bayes(self, theta, x, y, yerr, model__):
        """
        """
        # model = self.gaussian_model(x, theta, model)
        model = self.gaussian_model(self.axis, theta)
        sigma2 = yerr ** 2
        return -0.5 * np.sum((y - model) ** 2 / sigma2 + np.log(2 * np.pi * sigma2))

    def log_prior(self, theta, model):
        A_min = 0  # 1e-19
        A_max = 1.  # 1e-15
        x_min = 14700
        x_max = 15400
        sigma_min = 0
        sigma_max = 10
        for model_num in range(len(model)):
            params = theta[model_num * 3:(model_num + 1) * 3]
        within_bounds = True  # Boolean to determine if parameters are within bounds
        for ct, param in enumerate(params):
            if ct % 3 == 0:  # Amplitude parameter
                if param > A_min and param < A_max:
                    pass
                else:
                    within_bounds = False  # Value not in bounds
                    break
            if ct % 3 == 1:  # velocity parameter
                if param > x_min and param < x_max:
                    pass
                else:
                    within_bounds = False  # Value not in bounds
                    break
            if ct % 3 == 2:  # sigma parameter
                if param > sigma_min and param < sigma_max:
                    pass
                else:
                    within_bounds = False  # Value not in bounds
                    break
        if within_bounds:
            return 0.0
        else:
            return -np.inf
        # A_,x_,sigma_ = theta
        # if A_min < A_ < A_max and x_min < x_ < x_max and sigma_min < sigma_ < sigma_max:
        #    return 0.0#np.log(1/((t_max-t_min)*(rp_max-rp_min)*(b_max-b_min)))
        # return -np.inf

    def log_probability(self, theta, x, y, yerr, model):
        lp = self.log_prior(theta, model)
        if not np.isfinite(lp):
            return -np.inf
        return lp + self.log_likelihood_bayes(theta, x, y, yerr, model)

    def plot(self):
        """
        Plot initial spectrum and fitted spectrum

        """
        plt.clf()
        plt.plot(self.axis, self.sky, label='Spectrum')
        plt.plot(self.axis, self.fit_vector, label='Fit')
        plt.legend()
        plt.show()
        return None

    def check_lines(self):
        """
        This function checks to see that the lines provided are in the available options
        Return:
            Nothing if the user provides appropriate lines
            Else it will throw an error
        """
        if set(self.lines).issubset(self.line_dict):
            pass
        else:
            raise Exception('Please submit a line name in the available list: \n {}'.format(self.line_dict.keys()))

    def check_fitting_model(self):
        """
        This function checks to see that the model provided is in the available options
        Return:
            Nothing if the user provides an appropriate fitting model
            Else it will throw an error
        """
        if self.model_type in self.available_functions:
            pass
        else:
            print(self.model_type)
            raise Exception(
                'Please submit a fitting function name in the available list: \n {}'.format(self.available_functions))



from astropy.io import fits
import h5py
import os
from astropy.wcs import WCS
from tqdm import tqdm
import numpy as np
import keras
import pyregion
import time
from joblib import Parallel, delayed
from LuciFit import Fit
import matplotlib.pyplot as plt


class Luci():
    """
    This is the primary class for the general purpose line fitting code LUCI. This contains
    all io/administrative functionality. The fitting functionality can be found in the
    Fit class (Lucifit.py).
    """
    def __init__(self, cube_path, output_dir, object_name, redshift, ref_spec=None, model_ML_name=None):
        """
        Initialize our Luci class -- this acts similar to the SpectralCube class
        of astropy or spectral-cube.
        Args:
            cube_path: Full path to hdf5 cube with the hdf5 extension (e.x. '/user/home/M87.hdf5')
            output_dir: Full path to output directory
            object_name: Name of the object to fit. This is used for naming purposes. (e.x. 'M87')
            redshift: Redshift to the object. (e.x. 0.00428)
            ref_spec: Name of reference spectrum for machine learning algo (e.x. 'Reference-Spectrum-R5000')
            model_ML_name: Name of pretrained machine learning model
        """
        self.cube_path = cube_path
        self.output_dir = output_dir+'/Luci'
        if not os.path.exists(self.output_dir):
            os.mkdir(self.output_dir)
        self.object_name = object_name
        self.redshift = redshift
        self.ref_spec = ref_spec+'.fits'
        if model_ML_name != None or '':
            self.model_ML = keras.models.load_model(model_ML_name)
        else:
            self.model_ML = model_ML_name
        self.quad_nb = 0  # Number of quadrants in Hdf5
        self.dimx = 0  # X dimension of cube
        self.dimy = 0  # Y dimension of cube
        self.dimz = 0  # Z dimension of cube
        self.cube_final = None  # Complete data cube
        self.cube_binned = None  # Binned data cube
        self.header = None
        self.deep_image = None
        self.spectrum_axis = None
        self.wavenumbers_syn = None
        self.hdr_dict = None
        self.read_in_cube()
        self.spectrum_axis_func()
        self.read_in_reference_spectrum()


    def get_quadrant_dims(self, quad_number):
        """
        Calculate the x and y limits of a given quadrant in the HDF5 file. The
        data cube is saved in 9 individual arrays in the original HDF5 cube. This
        function gets the bouunds for each quadrant.
        Args:
            quad_number: Quadrant Number
        Return:
            x_min, x_max, y_min, y_max: Spatial bounds of quadrant
        """
        div_nb = int(np.sqrt(self.quad_nb))
        if (quad_number < 0) or (quad_number > self.quad_nb - 1):
            raise StandardError("quad_number out of bounds [0," + str(self.quad_nb- 1) + "]")
            return "SOMETHING FAILED"

        index_x = quad_number % div_nb
        index_y = (quad_number - index_x) / div_nb

        x_min = index_x * np.floor(self.dimx / div_nb)
        if (index_x != div_nb - 1):
            x_max = (index_x  + 1) * np.floor(self.dimx / div_nb)
        else:
            x_max = self.dimx

        y_min = index_y * np.floor(self.dimy / div_nb)
        if (index_y != div_nb - 1):
            y_max = (index_y  + 1) * np.floor(self.dimy / div_nb)
        else:
            y_max = self.dimy
        return int(x_min), int(x_max), int(y_min), int(y_max)


    def update_header(self, file):
        """
        Create a standard WCS header from the HDF5 header
        Args:
            file: hdf5 File object containing HDF5 file
        """
        hdr_dict = {}
        header_cols = [str(val[0]).replace("'b",'').replace("'", "").replace("b",'') for val in list(file['header'][()])]
        header_vals = [str(val[1]).replace("'b",'').replace("'", "").replace("b",'') for val in list(file['header'][()])]
        header_types = [val[3] for val in list(file['header'][()])]
        for header_col, header_val, header_type in zip(header_cols,header_vals, header_types):
            if 'bool' in str(header_type):
                hdr_dict[header_col] = bool(header_val)
            if 'float' in str(header_type):
                hdr_dict[header_col] = float(header_val)
            if 'int' in str(header_type):
                hdr_dict[header_col] = int(header_val)
            #print(header_type)
            else:
                try:
                    hdr_dict[header_col] = float(header_val)
                except:
                    hdr_dict[header_col] = str(header_val)
        hdr_dict['CTYPE3'] = 'WAVE-SIP'
        hdr_dict['CUNIT3'] = 'm'
        # Make WCS
        wcs_data = WCS(hdr_dict, naxis=2)
        self.header = wcs_data.to_header()
        self.hdr_dict = hdr_dict


    def create_deep_image(self, output_name=None):
        """
        Create deep image fits file of the cube. This takes the cube and averages
        the spectral axis. Then the deep image is saved as a fits file with the following
        naming convention: output_dir+'/'+object_name+'_deep.fits'
        """
        hdu = fits.PrimaryHDU()
        self.deep_image = np.mean(self.cube_final, axis=2)
        if output_name == None:
            output_name = self.output_dir+'/'+self.object_name+'_deep.fits'
        fits.writeto(output_name, self.deep_image.T, self.header, overwrite=True)

    def read_in_cube(self):
        """
        Function to read the hdf5 data into a 3d numpy array (data cube). We also
        translate the header to standard wcs format by calling the update_header function.
        """
        print('Reading in data...')
        file =  h5py.File(self.cube_path+'.hdf5', 'r')  # Read in file
        self.quad_nb = file.attrs['quad_nb']  # Get the number of quadrants
        self.dimx = file.attrs['dimx']  # Get the dimensions in x
        self.dimy = file.attrs['dimy']  # Get the dimensions in y
        self.dimz = file.attrs['dimz']  # Get the dimensions in z (spectral axis)
        self.cube_final = np.zeros((self.dimx, self.dimy, self.dimz))  # Complete data cube
        for iquad in tqdm(range(self.quad_nb)):
            xmin,xmax,ymin,ymax = self.get_quadrant_dims(iquad)
            iquad_data = file['quad00%i'%iquad]['data'][:]  # Save data to intermediate array
            iquad_data[(np.isfinite(iquad_data) == False)]= 1e-22 # Modifs
            iquad_data[(iquad_data < -1e-16)]= -1e-22 # Modifs
            iquad_data[(iquad_data > 1e-9)]= 1e-22 # Modifs
            self.cube_final[xmin:xmax, ymin:ymax, :] = iquad_data  # Save to correct location in main cube
        self.update_header(file)

    def spectrum_axis_func(self):
        """
        Create the x-axis for the spectra. We must construct this from header information
        since each pixel only has amplitudes of the spectra at each point.
        """
        len_wl = self.cube_final.shape[2]  # Length of Spectral Axis
        start = self.hdr_dict['CRVAL3']  # Starting value of the spectral x-axis
        end = start + (len_wl)*self.hdr_dict['CDELT3']  # End
        #step = hdr_dict['CDELT3']  # Step size
        self.spectrum_axis = np.array(np.linspace(start, end, len_wl)*(self.redshift+1), dtype=np.float32)  # Apply redshift correction


    def read_in_reference_spectrum(self):
        """
        Read in the reference spectrum that will be used in the machine learning
        algorithm to interpolate the true spectra so that they
        wil all have the same size (required for our CNN). The reference spectrum
        will be saved as self.wavenumbers_syn [cm-1].
        """
        ref_spec = fits.open(self.ref_spec)[1].data
        channel = []
        counts = []
        for chan in ref_spec:  # Only want SN3 region
            channel.append(chan[0])
            counts.append(np.real(chan[1]))
        min_ = np.argmin(np.abs(np.array(channel)-14700))
        max_ = np.argmin(np.abs(np.array(channel)-15600))
        self.wavenumbers_syn = np.array(channel[min_:max_], dtype=np.float32)

    def fit_entire_cube(self, lines, fit_function):
        """
        Fit the entire cube (all spatial dimensions)
        Args:
            lines: Lines to fit (e.x. ['Halpha', 'NII6583'])
            fit_function: Fitting function to use (e.x. 'gaussian')
        """
        x_min = 0
        x_max = self.cube_final.shape[0]
        y_min = 0
        y_max = self.cube_final.shape[1]
        self.fit_cube(lines, fit_function, x_min, x_max, y_min, y_max)


    def bin_cube(self, binning, x_min, x_max, y_min, y_max):
        """
        Function to bin cube into bin x bin sub cubes
        Args:
            binning: Size of binning (equal in x and y direction)
        Return:
            Binned cubed called self.cube_binned and new spatial limits
        """
        x_shape_new = int((x_max-x_min)/binning)
        y_shape_new = int((y_max-y_min)/binning)
        binned_cube = np.zeros((x_shape_new, y_shape_new, self.cube_final.shape[2]))
        for i in range(x_shape_new):
            for j in range(y_shape_new):
                summed_spec = self.cube_final[x_min+int(i*binning):x_min+int((i+1)*binning), y_min+int(j*binning):y_min+int((j+1)*binning), :]
                #print(summed_spec[-1])
                summed_spec = np.nansum(summed_spec, axis=0)
                #print(summed_spec[-1])
                summed_spec = np.nansum(summed_spec, axis=0)
                #print(summed_spec)
                binned_cube[i,j] = summed_spec[:]
        self.header_binned = self.header
        self.header_binned['CRVAL1'] = self.header_binned['CRVAL1']/binning
        self.header_binned['CRVAL2'] = self.header_binned['CRVAL2']/binning
        self.header_binned['CDELT1'] = self.header_binned['CDELT1']/binning
        self.header_binned['CDELT2'] = self.header_binned['CDELT2']/binning
        #print(binned_cube.shape)
        self.cube_binned = binned_cube / (binning**2)
        #return binned_cube / (binning**2)


    def save_fits(self, lines, velocity_fits, broadening_fits, ampls_fits, flux_fits, chi2_fits, header, output_name, binning):
        if binning is not None:
            output_name = output_name + "_" + str(binning)
        fits.writeto(output_name+'_velocity.fits', velocity_fits.T, header, overwrite=True)
        fits.writeto(output_name+'_broadening.fits', broadening_fits.T, header, overwrite=True)
        for ct,line_ in enumerate(lines):  # Step through each line to save their individual amplitudes
            fits.writeto(output_name+'_'+line_+'_Amplitude.fits', ampls_fits[:,:,ct].T, header, overwrite=True)
            fits.writeto(output_name+'_'+line_+'_Flux.fits', flux_fits[:,:,ct].T, header, overwrite=True)
        fits.writeto(output_name+'_Chi2.fits', chi2_fits.T, header, overwrite=True)


    def fit_cube(self, lines, fit_function, x_min, x_max, y_min, y_max, bkg=None, binning=None, bayes_bool=False, output_name=None):
        """
        Primary fit call to fit rectangular regions in the data cube. This wraps the
        LuciFits.FIT().fit() call which applies all the fitting steps. This also
        saves the velocity and broadening fits files.
        Args:
            lines: Lines to fit (e.x. ['Halpha', 'NII6583'])
            fit_function: Fitting function to use (e.x. 'gaussian')
            x_min: Lower bound in x
            x_max: Upper bound in x
            y_min: Lower bound in y
            y_max: Upper bound in y
            bkg: Background Spectrum (1D numpy array; default None)
            binning:  Value by which to bin (default None)
            bayes_bool: Boolean to determine whether or not to run Bayesian analysis (default None)
            output_name: User defined output path/name (default None)
        Return:
            Velocity and Broadening arrays (2d). Also return amplitudes array (3D).
        """
        # Initialize fit solution arrays
        if binning != None:
            self.bin_cube(binning, x_min, x_max, y_min, y_max)
            #x_min = int(x_min/binning) ; y_min = int(y_min/binning) ; x_max = int(x_max/binning) ;  y_max = int(y_max/binning)
            x_max = int((x_max-x_min)/binning) ;  y_max = int((y_max-y_min)/binning)
            x_min = 0 ; y_min = 0
        velocity_fits = np.zeros((x_max-x_min, y_max-y_min), dtype=np.float32)
        broadening_fits = np.zeros((x_max-x_min, y_max-y_min), dtype=np.float32)
        chi2_fits = np.zeros((x_max-x_min, y_max-y_min), dtype=np.float32)
        # First two dimensions are the X and Y dimensions.
        #The third dimension corresponds to the line in the order of the lines input parameter.
        ampls_fits = np.zeros((x_max-x_min, y_max-y_min, len(lines)), dtype=np.float32)
        flux_fits = np.zeros((x_max-x_min, y_max-y_min, len(lines)), dtype=np.float32)
        if output_name == None:
            output_name = self.output_dir+'/'+self.object_name
        for i in tqdm(range(x_max-x_min)):
            x_pix = x_min + i
            vel_local = []
            broad_local = []
            ampls_local = []
            flux_local = []
            chi2_local = []
            for j in range(y_max-y_min):
                y_pix = y_min+j
                if binning is not None:
                    sky = self.cube_binned[y_pix, x_pix, :]
                else:
                    sky = self.cube_final[y_pix, x_pix, :]
                if bkg is not None:
                    sky -= bkg  # Subtract background spectrum
                good_sky_inds = [~np.isnan(sky)]  # Clean up spectrum
                sky = sky[good_sky_inds]
                axis = self.spectrum_axis[good_sky_inds]
                # Call fit!
                fit = Fit(sky, axis, self.wavenumbers_syn, fit_function, lines,
                        self.model_ML, Plot_bool = False, bayes_bool=bayes_bool)
                fit_dict = fit.fit()
                # Save local list of fit values
                vel_local.append(fit_dict['velocity'])
                broad_local.append(fit_dict['broadening'])
                ampls_local.append(fit_dict['amplitudes'])
                flux_local.append(fit_dict['fluxes'])
                chi2_local.append(fit_dict['chi2'])
            # Update global array of fit values
            velocity_fits[i] = vel_local
            broadening_fits[i] = broad_local
            ampls_fits[i] = ampls_local
            flux_fits[i] = flux_local
            chi2_fits[i] = chi2_local
        # Write outputs (Velocity, Broadening, and Amplitudes)
        if binning is not None:
            self.save_fits(lines, velocity_fits, broadening_fits, ampls_fits, flux_fits, chi2_fits, self.header, output_name, binning)
            fits.writeto(output_name+'_velocity.fits', velocity_fits.T,  self.header, overwrite=True)
        else:
            self.save_fits(lines, velocity_fits, broadening_fits, ampls_fits, flux_fits, chi2_fits, self.header, output_name, binning)
        return velocity_fits, broadening_fits, flux_fits, chi2_fits

        #n_threads = 1
        #for i in range(x_max-x_min):
            #SNR_calc(VEL, BROAD, i)
        #Parallel(n_jobs=n_threads, backend="threading", batch_size=int((x_max-x_min)/n_threads))(delayed(SNR_calc)(VEL, BROAD, i) for i in range(x_max-x_min));
        #Parallel(n_jobs=n_threads, backend="threading")(delayed(SNR_calc)(VEL, BROAD, i) for i in tqdm(range(x_max-x_min)));

    def fit_region(self, lines, fit_function, region, bkg= None, bayes_bool=False, output_name=None):
        """
        Fit the spectrum in a region. This is an extremely similar command to fit_cube except
        it works for ds9 regions. We first create a mask from the ds9 region file. Then
        we step through the cube and only fit the unmasked pixels. Although this may not
        be the most efficient method, it does ensure the fidelity of the wcs system.
        Args:
            lines: Lines to fit (e.x. ['Halpha', 'NII6583'])
            fit_function: Fitting function to use (e.x. 'gaussian')
            region: Name of ds9 region file (e.x. 'region.reg'). You can also pass a boolean mask array.
            bkg: Background Spectrum (1D numpy array; default None)
            bayes_bool: Boolean to determine whether or not to run Bayesian analysis
            output_name: User defined output path/name
        Return:
            Velocity and Broadening arrays (2d). Also return amplitudes array (3D).
        """
        # Create mask
        if '.reg' in region:
            shape = (2048, 2064)#(self.header["NAXIS1"], self.header["NAXIS2"])  # Get the shape
            r = pyregion.open(region).as_imagecoord(self.header)  # Obtain pyregion region
            mask = r.get_mask(shape=shape)  # Calculate mask from pyregion region
        else:
            mask = np.load(region)
        # Set spatial bounds for entire cube
        x_min = 0
        x_max = self.cube_final.shape[0]
        y_min = 0
        y_max = self.cube_final.shape[1]
        # Initialize fit solution arrays
        velocity_fits = np.zeros((x_max-x_min, y_max-y_min), dtype=np.float32)
        broadening_fits = np.zeros((x_max-x_min, y_max-y_min), dtype=np.float32)
        if len(region.split('/')) > 1:  # If region file is a path, just keep the name for output purposes
            region = region.split('/')[-1]
        if output_name == None:
            output_name = self.output_dir+'/'+self.object_name+'_'+region.split('.')[0]
        # First two dimensions are the X and Y dimensions.
        #The third dimension corresponds to the line in the order of the lines input parameter.
        ampls_fits = np.zeros((x_max-x_min, y_max-y_min, len(lines)), dtype=np.float32)
        flux_fits = np.zeros((x_max-x_min, y_max-y_min, len(lines)), dtype=np.float32)
        for i in tqdm(range(x_max-x_min)):
            x_pix = x_min + i
            vel_local = []
            broad_local = []
            ampls_local = []
            flux_local = []
            for j in range(y_max-y_min):
                y_pix = y_min+j
                # Check if pixel is in the mask or not
                # If so, fit as normal. Else, set values to zero
                if mask[x_pix, y_pix] == True:
                    sky = self.cube_final[x_pix, y_pix, :]
                    if bkg is not None:
                        sky -= bkg
                    good_sky_inds = [~np.isnan(sky)]  # Clean up spectrum
                    sky = sky[good_sky_inds]
                    axis = self.spectrum_axis[good_sky_inds]
                    # Call fit!
                    fit = Fit(sky, axis, self.wavenumbers_syn, fit_function, lines,
                            self.model_ML, Plot_bool = False, bayes_bool=bayes_bool)
                    fit_dict = fit.fit()
                    # Save local list of fit values
                    vel_local.append(fit_dict['velocity'])
                    broad_local.append(fit_dict['broadening'])
                    ampls_local.append(fit_dict['amplitudes'])
                    flux_local.append(fit_dict['fluxes'])
                else:
                    vel_local.append(0)
                    broad_local.append(0)
                    ampls_local.append([0]*len(lines))
                    flux_local.append([0]*len(lines))
            # Update global array of fit values
            velocity_fits[i] = vel_local
            broadening_fits[i] = broad_local
            ampls_fits[i] = ampls_local
            flux_fits[i] = flux_local
            # Write outputs (Velocity, Broadening, and Amplitudes)
        fits.writeto(output_name+'_velocity.fits', velocity_fits.T, self.header, overwrite=True)
        fits.writeto(output_name+'_broadening.fits', broadening_fits.T, self.header, overwrite=True)
        for ct,line_ in enumerate(lines):  # Step through each line to save their individual amplitudes
            fits.writeto(output_name+line_+'_Amplitude.fits', ampls_fits[:,:,ct].T, self.header, overwrite=True)
            fits.writeto(output_name+line_+'_Flux.fits', flux_fits[:,:,ct].T, self.header, overwrite=True)

        return velocity_fits, broadening_fits, flux_fits


    def extract_spectrum_region(self, region, mean=False):
        """
        Extract spectrum in region. This is primarily used to extract background regions.
        The spectra in the region are summed and then averaged (if mean is selected).
        Using the 'mean' argument, we can either calculate the total summed spectrum (False)
        or the averaged spectrum for background spectra (True).
        Args:
            region: Name of ds9 region file (e.x. 'region.reg'). You can also pass a boolean mask array.
            mean: Boolean to determine whether or not the mean spectrum is taken. This is used for calculating background spectra.
        Return:
            X-axis and spectral axis of region.
        """
        # Create mask
        if '.reg' in region:
            shape = (2048, 2064)#(self.header["NAXIS1"], self.header["NAXIS2"])  # Get the shape
            r = pyregion.open(region).as_imagecoord(self.header)  # Obtain pyregion region
            mask = r.get_mask(shape=shape)  # Calculate mask from pyregion region
        else:
            mask = region
        # Set spatial bounds for entire cube
        x_min = 0
        x_max = self.cube_final.shape[0]
        y_min = 0
        y_max = self.cube_final.shape[1]
        integrated_spectrum = np.zeros(self.cube_final.shape[2])
        spec_ct = 0
        for i in tqdm(range(x_max-x_min)):
            x_pix = x_min + i
            for j in range(y_max-y_min):
                y_pix = y_min+j
                # Check if pixel is in the mask or not
                if mask[x_pix, y_pix] == True:
                    integrated_spectrum += self.cube_final[y_pix, x_pix, :]
                    spec_ct += 1
                else:
                    pass
        if mean == True:
            integrated_spectrum /= spec_ct
        return self.spectrum_axis, integrated_spectrum


    def create_snr_map(self, x_min=0, x_max=2048, y_min=0, y_max=2064, method=1):
        """
        Create signal-to-noise ratio (SNR) map of a given region. If no bounds are given,
        a map of the entire cube is calculated.
        Args:
            x_min: Minimal X value (default 0)
            x_max: Maximal X value (default 2048)
            y_min: Minimal Y value (default 0)
            y_max: Maximal Y value (default 2064)
            method: Method used to calculate SNR (default 1; options 1 or 2)
        Return:
            snr_map: Signal-to-Noise ratio map
        """
        # Calculate bounds for SNR calculation
        # Step through spectra
        SNR = np.zeros((x_max-x_min, y_max-y_min), dtype=np.float32)
        start = time.time()
        def SNR_calc(SNR, i):
            x_pix = x_min + i
            snr_local = []
            for j in range(y_max-y_min):
                y_pix = y_min+j
                # Calculate SNR
                flux = np.real(self.cube_final[x_pix, y_pix])
                flux = flux[np.where(flux != 0.0)]
                n = len(flux)
                if method == 1:
                    signal = np.max(flux)-np.median(flux)
                    noise = np.std(flux)
                    snr = float(signal / np.sqrt(noise))
                    snr = snr/(np.sqrt(np.mean(flux)))
                else:
                    # Select spectral region around Halpha and NII complex
                    min_ = np.argmin(np.abs(np.array(self.spectrum_axis)-15150))
                    max_ = np.argmin(np.abs(np.array(self.spectrum_axis)-15300))
                    flux_in_region = np.sum(self.cube_final[x_pix, y_pix, min_:max_])
                    # Select distance region
                    min_ = np.argmin(np.abs(np.array(self.spectrum_axis)-14800))
                    max_ = np.argmin(np.abs(np.array(self.spectrum_axis)-14850))
                    std_out_region = np.std(self.cube_final[x_pix, y_pix, min_:max_])
                    snr = float(flux_in_region/np.sqrt(std_out_region))
                    snr = snr#/(np.sqrt(np.mean(flux_in_region)))
                snr_local.append(snr)
            SNR[i] = snr_local
        n_threads = 16
        Parallel(n_jobs=n_threads, backend="threading", batch_size=int((x_max-x_min)/n_threads))(delayed(SNR_calc)(SNR,i) for i in tqdm(range(x_max-x_min)));
        end = time.time()
        # Save
        fits.writeto(self.output_dir+'/'+self.object_name+'_SNR.fits', SNR.T, self.header, overwrite=True)



"""
Set of functions required for Bayesian analysis using LUCI
"""
from LuciFit import Fit


def log_likelihood(theta, x, y, yerr, model):
    """
    theta - list of parameters for gaussian fit
    """
    #if model == 1:
    #    A_,B_,x_,sigma_ = theta
    #    model = gaussian_model(x, A_, B_, x_, sigma_)
    #elif model == 2:
    #    A_,B_,x_,sigma_, A2_, x2_, sigma2_ = theta
    #    model = gaussian_model2(x, A_, B_, x_, sigma_, A2_, x2_, sigma2_)
    model = Fit.gaussian_model(x, theta, model)
    sigma2 = yerr ** 2
    return -0.5 * np.sum((y - model) ** 2 / sigma2 + np.log(2*np.pi*sigma2))

def log_prior(theta, model):
    A_min = 0#1e-19
    A_max = 1.#1e-15
    x_min = 14700
    x_max = 15400
    sigma_min = 0
    sigma_max = 10
    for model_num in range(len(model)):
        params = theta[model_num*3:(model_num+1)*3]
    within_bounds = True  # Boolean to determine if parameters are within bounds
    for ct, param in enumerate(params):
        if ct%3 == 0:  # Amplitude parameter
            if param > A_min and param < A_max:
                pass
            else:
                within_bounds = False  # Value not in bounds
                break
        if ct%3 == 1:  # velocity parameter
            if param > x_min and param < x_max:
                pass
            else:
                within_bounds = False  # Value not in bounds
                break
        if ct%3 == 2:  # sigma parameter
            if param > sigma_min and param < sigma_max:
                pass
            else:
                within_bounds = False  # Value not in bounds
                break
    if within_bounds:
        return 0.0
    else:
        return -np.inf
    #A_,x_,sigma_ = theta
    #if A_min < A_ < A_max and x_min < x_ < x_max and sigma_min < sigma_ < sigma_max:
    #    return 0.0#np.log(1/((t_max-t_min)*(rp_max-rp_min)*(b_max-b_min)))
    #return -np.inf



def log_probability(theta, x, y, yerr, model):
    lp = log_prior(theta, model)
    if not np.isfinite(lp):
        return -np.inf
    return lp + log_likelihood(theta, x, y, yerr, model)



from setuptools import setup, find_packages

VERSION = '0.0.1'
DESCRIPTION = 'General Purpose Emission Line Fitting Package'
LONG_DESCRIPTION = 'General purpose emission line fitting package developed for the SITELLE IFU'

# Setting up
setup(
       # the name must match the folder name 'verysimplemodule'
        name="luci-fit",
        version=VERSION,
        author="Carter Rhea",
        author_email="carter.rhea@umontreal.ca",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        packages=find_packages(),
        install_requires=[], # add any additional packages that
        # needs to be installed along with your package. Eg: 'caer'
        keywords=['python', 'luci-fit'],
        classifiers= [
            "Programming Language :: Python :: 3",
        ]
)
